/*
 *  ISAS CVS Group 1993.11.15   presented by K.Suzuki
 * 
 *      create_image.c  ( xpmfview )
 *
 */

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#define NOT_ALLOC -1
#define QUANT     128

extern ShowMessage;

static int      dither[8][8] = {
	 0, 32,  8, 40,  2, 34, 10, 42,
	48, 16, 56, 24, 50, 18, 58, 26,
	12, 44,  4, 36, 14, 46,  6, 38,
	50, 28, 52, 20, 62, 30, 54, 22,
	 3, 35, 11, 43,  1, 33,  9, 41,
	51, 19, 59, 27, 49, 17, 57, 25,
	15, 47,  7, 39, 13, 45,  5, 37,
	63, 21, 55, 23, 61, 29, 53, 21
};



static XImage  *
CreatePseudoGray( disp, visual, image, width, height )
     Display        *disp;
     Visual         *visual;
     unsigned char  *image;
     int             width, height;
{
  unsigned char  *gimage, *gim;
  int             i, val, ncolors;
  short           pixel_table[QUANT];
  XColor          color;

  for ( i = 0; i < QUANT; i++ )
    pixel_table[i] = NOT_ALLOC;

  gimage = gim = (unsigned char *)malloc( width * height );

  for( ncolors = i = 0; i < height * width; i++ ){
    val = *image++ >> 1;

    if( pixel_table[val] == NOT_ALLOC ){
      color.red = color.green = color.blue = 512 * val;
      if( !XAllocColor( disp, DefaultColormap(disp, 0), &color ) ){
	fprintf( stderr, "too many colors %d!\n", ncolors );
	exit( 1 );
      }
      ncolors++;
      pixel_table[val] = color.pixel;
    }
    *gim++ = pixel_table[val];
  }
  if( ShowMessage )
    fprintf( stderr, "%d colors used.\n", ncolors );
  return XCreateImage( disp, visual, 8, ZPixmap, 0, gimage,
		       width, height, 8, 0 );
}


static XImage  *
CreateDither( disp, visual, image, width, height )
     Display        *disp;
     Visual         *visual;
     unsigned char  *image;
     int             width, height;
{
  unsigned char  *dimage, *dim;
  XColor          color;
  int             i, j, k, r, g, b, rm, gm, bm, dith, ncolors;
  short           pixel_table[5][9][5];

  for( i = 0; i < 5; i++ ){
    for( j = 0; j < 9; j++ ){
      for( k = 0; k < 5; k++ ){
	pixel_table[i][j][k] = NOT_ALLOC;
      }
    }
  }
  dimage = dim = (unsigned char *)malloc( width * height );
  for( ncolors = i = 0; i < height; i++) {
    for( j = 0; j < width; j++ ){
      dith = dither[i % 8][j % 8];
      r = *image++; g = *image++; b = *image++;
      rm = r % 64;  gm = g % 32;  bm = b % 64;
      r /= 64;      g /= 32;      b /= 64;
      if( dith < rm )	  r++;
      if( dith < 2 * gm ) g++;
      if( dith < bm ) 	  b++;
      if( pixel_table[r][g][b] == NOT_ALLOC ) {
	color.red  = 16383 * r;
	color.green = 8191 * g;
	color.blue = 16383 * b;
	if( !XAllocColor( disp, DefaultColormap(disp, 0), &color ) ){
	  fprintf( stderr, "too many colors %d !\n", ncolors );
	  exit( 1 );
	}
	pixel_table[r][g][b] = color.pixel;
	ncolors++;
      }
      *dim++ = pixel_table[r][g][b];
    }
  }
  if( ShowMessage )
    fprintf( stderr, "%d colors used.\n", ncolors );
  return XCreateImage( disp, visual, 8, ZPixmap, 0, dimage,
		       width, height, 8, 0);
}


XImage *
CreatePseudoColorImage( disp, visual, image, width, height, depth )
     Display        *disp;
     Visual         *visual;
     unsigned char  *image;
     int             width, height, depth;
{
  switch( depth ){
  case 1:
    return XCreateImage( disp, visual, 1, XYBitmap, 0, image,
			 width, height, 8, 0 );
  case 8:
    return CreatePseudoGray( disp, visual, image, width, height );
  case 24:
    return CreateDither( disp, visual, image, width, height );
  }
}


static XImage  *
CreateTrueGray( disp, visual, image, width, height )
     Display        *disp;
     Visual         *visual;
     unsigned char  *image;
     int             width, height;
{
  int             i;
  unsigned char  *im, *newimage;

  im = newimage = (unsigned char *)malloc( width * height * 4 );
  for( i = 0; i < width * height; i++ ){
    *im++;
    *im++ = *image;
    *im++ = *image;
    *im++ = *image++;
  }
  return XCreateImage( disp, visual, 24, ZPixmap, 0, newimage,
		       width, height, 32, 0);
}


static XImage  *
CreateTrueColor( disp, visual, image, width, height )
     Display        *disp;
     Visual         *visual;
     unsigned char  *image;
     int             width, height;
{
  int             i;
  unsigned char  *im, *newimage;

  im = newimage = (unsigned char *)malloc( width * height * 4 );
  for( i = 0; i < width * height; i++ ){
    *im++;
    *im++ = *image++;
    *im++ = *image++;
    *im++ = *image++;
  }
  return XCreateImage( disp, visual, 24, ZPixmap, 0, newimage,
		       width, height, 32, 0);
}


XImage         *
CreateTrueColorImage( disp, visual, image, width, height, depth )
     Display        *disp;
     Visual         *visual;
     unsigned char  *image;
     int             width, height, depth;
{
  switch( depth ){
  case 1:
    return XCreateImage( disp, visual, 1, XYBitmap, 0, image,
			 width, height, 8, 0);
  case 8:
    return CreateTrueGray( disp, visual, image, width, height );
  case 24:
    return CreateTrueColor( disp, visual, image, width, height );
  }
}


XImage         *
CreateStaticGrayImage( disp, visual, image, width, height, depth )
     Display        *disp;
     Visual         *visual;
     unsigned char  *image;
     int             width, height, depth;
{
  unsigned char  *bim, *bimage;
  int             i, j, r, g, b, bitmask, byte;

  if( depth == 1 )
    return XCreateImage( disp, visual, 1, ZPixmap, 0, image,
			 width, height, 8, 0 );

  bimage = bim = (unsigned char *)malloc( width * height / 8 );
  for( bitmask = 1 << 7, byte = 0, i = 0; i < height; i++ ){
    for( j = 0; j < width; j++ ){
      switch( depth ){
      case 24:
	r = *image++; g = *image++; b = *image++;
	g = (299 * r + 587 * g + 114 * b) / 1000;
	/*
	 * NTSC formula: comp.graphics FAQ, Jan. 92
	 * by Jef Poskanzer  jef@well.sf.ac.us
	 * apple!well!jef
	 */
	break;
      case 8:
	g = *image++;
      }
      g = g >> 2;
      if( dither[i % 8][j % 8] > g )
	byte |= bitmask;
      bitmask = bitmask >> 1;
      if( !bitmask ){
	bitmask = 1 << 7;
	*bim++ = byte;
	byte = 0;
      }
    }
  }
  return XCreateImage( disp, visual, 1, ZPixmap, 0, bimage,
		       width, height, 8, 0);
}
