/*
 *  pixel map format  library
 *                             programed by Kenya.Suzuki
 */

#include <stdio.h>
#include <malloc.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>

#define PMF_LIBRARY
#include <pmf.h>


/*
 *  static constants
 */
static char *magic_no[] = {
  PMF_MN_BITMAP, PMF_MN_GRAYMAP, PMF_MN_PIXELMAP, PMF_MN_ORIGINAL
};

static unsigned char bit_mask[] = {
  0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};

static unsigned char bit_reset[] = {
  0x7F, 0xBF, 0xDF, 0xEF, 0xF7, 0xFB, 0xFD, 0xFE
};

static char *error_mes[] = {
  "undefined error !!",
  "unsupported format is selected.",
  "file I/O error.",
  "read unsupported format file.",
  "call by illegal parameter."
};


/*
 *  librarys
 */
unsigned
PMFSize( pmf )
     PMF *pmf;
{
  unsigned size = 0;

  if( pmf != NULL ){
    switch( pmf->type ){
    case PMF_BITMAP:
      size = ( ( pmf->width + 7 ) / 8 ) * pmf->height * sizeof(PMFBitMap);
      break;
    case PMF_GRAYMAP:
      size = pmf->width * pmf->height * sizeof(PMFGrayMap);
      break;
    case PMF_PIXELMAP:
      size = pmf->width * pmf->height * sizeof(PMFPixelMap);
      break;
    }
  } else {
    PMFerrno = PMF_ERR_ILLEGAL;
  }
  return( size );
}

PMF *
PMFCreate( type, width, height )
     int type, width, height;
{
  PMF *pmf = NULL;
  char     *buf;
  unsigned  size;

  if(( 0 <= type )&&( type <= PMF_FORMAT_MAX )){
    if(( width > 0 )&&( height > 0 )){
      if(( pmf = (PMF *)malloc( sizeof(PMF) ) ) != NULL ){
	pmf->type   = type;
	pmf->width  = width;
	pmf->height = height;
	if( type == PMF_BITMAP )
	  pmf->maxval = 1;
	else if(( type == PMF_GRAYMAP )||( type == PMF_PIXELMAP ))
	  pmf->maxval = 255;

	size = PMFSize( pmf );

	if(( buf = malloc( size ) ) != NULL ){
	  pmf->pm.ptr = (unsigned char *)buf;
	} else {
	  PMFerrno = PMF_ERR_OS;
	  free( pmf );
	  pmf = NULL;
	}

      } else {
	PMFerrno = PMF_ERR_OS;
      }

    } else {
      PMFerrno = PMF_ERR_ILLEGAL;
    }
  } else {
    PMFerrno = PMF_ERR_TYPE;
  }
  return( pmf );
}

void
PMFErase( pmf )
     PMF *pmf;
{
  if( pmf != NULL ){
    free( pmf->pm.ptr );
    free( pmf );
  }
}

static void
_PMFperror( str, mes )
     char *str, *mes;
{
  if(( str != NULL )&&( *str != '\0' ))
    fprintf( stderr, "%s: %s\n", str, mes );
  else
    fprintf( stderr, "%s\n", mes );
}

void
PMFperror( str )
     char *str;
{
  if( PMFerrno != PMF_ERR_NONE ){
    if( PMFerrno  == PMF_ERR_OS )
      perror( str );
    else if( PMFerrno <= PMF_ERR_MAX ){
      _PMFperror( str, error_mes[PMFerrno - 1] );
    } else {
      _PMFperror( str, error_mes[0] );
    }
  }
}

static void
_PMFfix_color( pmf )
     PMF  *pmf;
{
  int      i, maxval;
  double   grad;
  unsigned size;
  unsigned char *ptr;
  unsigned char map[256];

  if( pmf->type != PMF_BITMAP ){
    size = PMFSize( pmf );

    maxval = pmf->maxval;
    grad  = 255.0 / (double)maxval;
    for( i = 0; i <= maxval; i++ ){
      map[i] = (PMFGrayMap)( grad * (double)i );
    }
    ptr = pmf->pm.ptr;
    for( ; size > 0; size-- ){
      *ptr = map[(int)*ptr];
      ptr++;
    }
    pmf->maxval = 255;
  }
}

static int
_PMFgetpara( fp )
     FILE *fp;
{
  int  ret = -1;
  int  flag = 0;
  int  c;
  unsigned char buf[128], *str;

  str = buf;
  while( 1 ){
    if(( c = fgetc( fp ) ) == EOF ){
      PMFerrno = PMF_ERR_FILEIO;
      goto RET;
    }
    if(( c == ' ' )||( c == '\t' )||( c == '\r' )||( c == '\n' )){
      if( flag ){
	*str = '\0';
	break;
      }
    } else if( c == '#' ){
      while( c != '\n' )
	if(( c = fgetc( fp ) ) == EOF ){
	  PMFerrno = PMF_ERR_FILEIO;
	  goto RET;
	}
    } else if( isdigit( c ) ){
      flag = 1;
      *str++ = c;
    } else {
      PMFerrno = PMF_ERR_FILEIO;
      goto RET;
    }
  }
  ret = atoi( buf );
RET:
  return( ret );
}

static PMF *
_PMFLoad( fp )
     FILE *fp;
{
  int  type, width, height, maxval;
  PMF  *pmf = NULL;
  unsigned size;
  char buf[2];

  if( fread( buf, sizeof(char), 2, fp ) != 2 ){
    PMFerrno = PMF_ERR_FILEIO;
    goto RET;
  }
  for( type = 0; type <= PMF_FORMAT_MAX; type++ ){
    if( strncmp( buf, magic_no[type], 2 ) == 0 )
      break;
  }
  if( type > PMF_FORMAT_MAX ){
    PMFerrno = PMF_ERR_UNSUPPORT;
    goto RET;
  }

  if(( width  = _PMFgetpara( fp ) ) < 0 )
    goto RET;
  if(( height = _PMFgetpara( fp ) ) < 0 )
    goto RET;

  if( type == PMF_BITMAP ){
    maxval = 1;
  } else {
    if(( maxval = _PMFgetpara( fp ) ) < 0 )
      goto RET;
  }

  if(( pmf = PMFCreate( type, width, height ) ) == NULL )
    goto RET;

  pmf->maxval = maxval;

  size = PMFSize( pmf );

  if( fread( pmf->pm.ptr, sizeof(char), size, fp ) != size ){
    PMFerrno = PMF_ERR_FILEIO;
    PMFErase( pmf );
    pmf = NULL;
    goto RET;
  }

  _PMFfix_color( pmf );

RET:
  return( pmf );
}

PMF *
PMFLoad( file )
     char *file;
{
  PMF  *pmf = NULL;
  FILE *fp;

  if( file == NULL ){
    pmf = _PMFLoad( stdin );

  } else if(( fp = fopen( file, "r" ) ) != NULL ){
    pmf = _PMFLoad( fp );

    if( fclose( fp ) == EOF ){
      PMFerrno = PMF_ERR_FILEIO;
      pmf = NULL;
    }
  } else {
    PMFerrno = PMF_ERR_FILEIO;
  }
  return( pmf );
}

static int
_PMFSave( fp, pmf, comment )
     FILE *fp;
     PMF  *pmf;
     char *comment;
{
  int ret = 1;
  unsigned size;

  if( fprintf( fp, "%s\n", magic_no[pmf->type] ) == EOF ){
    PMFerrno = PMF_ERR_FILEIO;
    goto RET;
  }

  if(( comment != NULL )&&( *comment != '\0' )){
    if( fprintf( fp, "# %s\n", comment ) == EOF ){
      PMFerrno = PMF_ERR_FILEIO;
      goto RET;
    }
  }

  if( fprintf( fp, "%d %d\n", pmf->width, pmf->height ) == EOF ){
    PMFerrno = PMF_ERR_FILEIO;
    goto RET;
  }

  if( pmf->type != PMF_BITMAP ){
    if( fprintf( fp, "%d\n", pmf->maxval ) == EOF ){
      PMFerrno = PMF_ERR_FILEIO;
      goto RET;
    }
  }

  size = PMFSize( pmf );

  if( fwrite( (char *)pmf->pm.ptr, sizeof(char), size ,fp ) != size ){
    PMFerrno = PMF_ERR_FILEIO;
    goto RET;
  }

  ret = 0;
RET:
  return( ret );
}

int
PMFSave( file, pmf, comment )
     char *file, *comment;
     PMF  *pmf;
{
  int ret = 1;
  FILE *fp;

  if( pmf != NULL ){
    if(( 0 <= pmf->type )&&( pmf->type <= PMF_FORMAT_MAX )){
      if( file == NULL ){
	ret = _PMFSave( stdout, pmf, comment );

      } else if(( fp = fopen( file, "w" ) ) != NULL ){
	ret = _PMFSave( fp, pmf, comment );

	if( fclose( fp ) == EOF ){
	  PMFerrno = PMF_ERR_FILEIO;
	  ret = 1;
	}
      } else {
	PMFerrno = PMF_ERR_FILEIO;
      }
    } else {
      PMFerrno = PMF_ERR_TYPE;
    }
  } else {
    PMFerrno = PMF_ERR_ILLEGAL;
  }
  return( ret );
}

int
PMFClear( pmf )
     PMF *pmf;
{
  int  ret = 0;
  unsigned       size;
  unsigned char *ptr;

  if( pmf != NULL ){
    size = PMFSize( pmf );
    ptr = pmf->pm.ptr;
    for( ; size > 0; size-- )
      *ptr++ = 0;
  } else {
    PMFerrno = PMF_ERR_ILLEGAL;
    ret = 1;
  }
  return( ret );
}

PMFPixel
PMFGetPixel( pmf, x, y )
     PMF  *pmf;
     int  x, y;
{
  PMFPixel  pix;
  int       width, mask, offset;
  PM_Ptr    pm;

  if( pmf != NULL ){
    width = pmf->width;
    if(( 0 <= x )&&( x < width )&&( 0 <= y )&&( y < pmf->height )){
      switch( pmf->type ){
      case PMF_BITMAP:
	mask   = x % 8;
	offset = ( x / 8 ) + y * (( width + 7 )/ 8);
	pm.bitmap = pmf->pm.bitmap + offset;
	if( *pm.bitmap & bit_mask[mask] != 0 )
	  pix.bit = 1;
	else
	  pix.bit = 0;
	break;
      case PMF_GRAYMAP:
	offset = x + y * width;
	pm.graymap = pmf->pm.graymap + offset;
	pix.gray = (long)*pm.graymap;
	break;
      case PMF_PIXELMAP:
	offset = x + y * width;
	pm.pixelmap = pmf->pm.pixelmap + offset;
	pix.pixel.x = 0;
	pix.pixel.r = pm.pixelmap->r;
	pix.pixel.g = pm.pixelmap->g;
	pix.pixel.b = pm.pixelmap->b;
	break;
      }
    } else {
      PMFerrno = PMF_ERR_ILLEGAL;
      pix.err = -1;
    }
  } else {
    PMFerrno = PMF_ERR_ILLEGAL;
    pix.err = -1;
  }
  return( pix );
}

int
PMFPutPixel( pmf, x, y, pix )
     PMF      *pmf;
     int       x, y;
     PMFPixel  pix;
{
  int       ret = 0;
  int       width, mask, offset;
  PM_Ptr    pm;

  if(( pmf != NULL )&&( pix.err >= 0 )){
    width = pmf->width;
    if(( 0 <= x )&&( x < width )&&( 0 <= y )&&( y < pmf->height )){
      switch( pmf->type ){
      case PMF_BITMAP:
	mask   = x % 8;
	offset = ( x / 8 ) + y * (( width + 7 )/ 8 );
	pm.bitmap = pmf->pm.bitmap + offset;
	if( pix.bit != 0 )
	  *pm.bitmap |= bit_mask[mask];
	else
	  *pm.bitmap &= bit_reset[mask];
	break;
      case PMF_GRAYMAP:
	offset = x + y * width;
	pm.graymap = pmf->pm.graymap + offset;
	*pm.graymap = (unsigned char)pix.gray;
	break;
      case PMF_PIXELMAP:
	offset = x + y * width;
	pm.pixelmap = pmf->pm.pixelmap + offset;
	pm.pixelmap->r = pix.pixel.r;
	pm.pixelmap->g = pix.pixel.g;
	pm.pixelmap->b = pix.pixel.b;
	break;
      }
    } else {
      PMFerrno = PMF_ERR_ILLEGAL;
      ret = 1;
    }
  } else {
    PMFerrno = PMF_ERR_ILLEGAL;
    ret = 1;
  }
  return( ret );
}
