/*
 *  pixel map format  definition file
 *
 *                             presented by Kenya.Suzuki
 */

#ifndef DEFINE_PMF
#define DEFINE_PMF

/*
 *  the kind of pixel map format
 */
#define PMF_BITMAP     0
#define PMF_GRAYMAP    1
#define PMF_PIXELMAP   2
#define PMF_ORIGINAL   3
#define PMF_FORMAT_MAX 3

/*
 *  the magic number of pixel map format
 */
#define PMF_MN_BITMAP   "P4"
#define PMF_MN_GRAYMAP  "P5"
#define PMF_MN_PIXELMAP "P6"
#define PMF_MN_ORIGINAL "PMF"

/*
 *  the definition of error number for pixel map format
 */
#define PMF_ERR_NONE       0
#define PMF_ERR_OS         1
#define PMF_ERR_TYPE       2
#define PMF_ERR_FILEIO     3
#define PMF_ERR_UNSUPPORT  4
#define PMF_ERR_ILLEGAL    5
#define PMF_ERR_MAX        5


/*
 *  the internal data type for pixel map format
 */
struct pmf_pixelmap {
  unsigned char r;
  unsigned char g;
  unsigned char b;
};

typedef unsigned char PMFBitMap;
typedef unsigned char PMFGrayMap;
typedef struct pmf_pixelmap PMFPixelMap;

union pm_ptr {
  PMFBitMap     *bitmap;
  PMFGrayMap    *graymap;
  PMFPixelMap   *pixelmap;
  unsigned char *ptr;
};

typedef union pm_ptr PM_Ptr;

struct pmf {
  short  type;
  short  width;
  short  height;
  short  maxval;
  PM_Ptr pm;
};

typedef struct pmf PMF;

union pmf_pixel {
  long  bit;
  long  gray;
  struct {
    char          x;
    unsigned char r;
    unsigned char g;
    unsigned char b;
  } pixel;
  long  err;
};

typedef union pmf_pixel PMFPixel;


/*
 *  the macros for pixel map format
 */
#define PMFType( pmf )     (pmf)->type
#define PMFWidth( pmf )    (pmf)->width
#define PMFHeight( pmf )   (pmf)->height
#define PMFPMPtr( pmf )    (pmf)->pm.ptr

#define PMFSetBit( pix, col )       (pix).bit = (col)
#define PMFSetGray( pix, col )      (pix).gray = (col)
#define PMFSetPixel( pix, r, g, b ) (pix).pixel.x = 0;\
				    (pix).pixel.r = (unsigned char)(r);\
				    (pix).pixel.g = (unsigned char)(g);\
				    (pix).pixel.b = (unsigned char)(b)


/*
 *  the library for pixel map format
 */
#ifdef PMF_LIBRARY
#define PMF_extern
#else
#define PMF_extern extern
#endif

#ifdef __GNUC__
PMF_extern PMF     *PMFCreate( int, int, int );
PMF_extern void     PMFErase( PMF * );
PMF_extern PMF     *PMFLoad( char * );
PMF_extern int      PMFSave( char *, PMF *, char * );
PMF_extern unsigned PMFSize( PMF * );
PMF_extern void     PMFperror( char * );
PMF_extern int      PMFClear( PMF * );
PMF_extern PMFPixel PMFGetPixel( PMF *, int, int );
PMF_extern int      PMFPutPixel( PMF *, int, int, PMFPixel );

#else
PMF_extern PMF     *PMFCreate();
PMF_extern void     PMFErase();
PMF_extern PMF     *PMFLoad();
PMF_extern int      PMFSave();
PMF_extern unsigned PMFSize();
PMF_extern void     PMFperror();
PMF_extern int      PMFClear();
PMF_extern PMFPixel PMFGetPixel();
PMF_extern int      PMFPutPixel();

#endif

#undef PMF_extern


#ifdef PMF_LIBRARY
#ifdef __GNUC__
extern PMF     *PMFCreate( int, int, int );
extern void     PMFErase( PMF * );
extern PMF     *PMFLoad( char * );

#else

#endif
#endif


/*
 *  extern variable
 */
#ifndef PMF_LIBRARY
extern int PMFerrno;

#else
int PMFerrno = PMF_ERR_NONE;

#endif

#endif

/**** End of pmf.h ****/
