/*
 *  ISAS CVS Group 1993.11.15   presented by K.Suzuki
 * 
 *         xpmfview.c
 * 
 */

#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <pmf.h>


static Display *disp;
static Window   win;
static Visual  *visual;
static XImage  *ximage;

static PMF     *pmf;

static int      depth = 24, width = 0, height = 0;
static char    *filename = NULL;
static char    *dispstr  = NULL;
Bool            ShowMessage = False;


static unsigned char table[256];


static void
usage( argv )
     char *argv[];
{
  char *str;

  if(( str = strrchr( argv[0], '/' ) ) != NULL )
    str++;
  else
    str = argv[0];
  fprintf( stderr, 
	  "usage: %s [-d display] [-v] [-rev] [filename]\n", str );
}


static void 
get_args( argc, argv )
     int   argc;
     char *argv[];
{
  int       i;
  unsigned  size;
  char     *str;
  unsigned char *ptr;
  FILE     *fp;
  Bool      reverse_mode = False;

  for( i = 1; i < argc; i++ ){
    str = argv[i];
    if(( *str == '-' )&&( *(str+1) != '-' )&&( *(str+1) != '\0' )){
      str++;
      switch( *str ){
      case 'd':
	if( ++i < argc )
	  dispstr = argv[i];
	break;
      case 'v':
	ShowMessage = True;
	break;
      default:
	if( strcmp( str, "rev" ) == 0 )
	  reverse_mode = True;
	else {
	  usage( argv );
	  exit( 1 );
	}
      }
    } else {
      filename = (char *)malloc( strlen( argv[i] ) + 1 );
      strcpy( filename, argv[i] );
    }
  }

  if(( pmf = PMFLoad( filename ) ) == NULL ){
    PMFperror( "xpmfview" );
    exit( 1 );
  }

  width  = PMFWidth( pmf );
  height = PMFHeight( pmf );
  switch( PMFType( pmf ) ){
  case PMF_BITMAP:
    depth  = 1;
    break;
  case PMF_GRAYMAP:
    depth  = 8;
    break;
  case PMF_PIXELMAP:
    depth  = 24;
    break;
  }

  if( reverse_mode ){
    if( ShowMessage )
      fprintf( stderr, "Reverse Mode.\n" );
    ptr = pmf->pm.ptr;
    for( size = PMFSize( pmf ); size > 0; size--, ptr++ )
      *ptr = *ptr ^ 0xFF;
  }
}


static void
bitmap_fit( disp, pmf )
     Display *disp;
     PMF     *pmf;
{
  int      i, d;
  unsigned size;
  unsigned char *ptr;
  
  if( PMFType( pmf ) == PMF_BITMAP ){
    /*
     *  bit reverse because of bit order diffence.
     */
    if( BitmapBitOrder( disp ) == LSBFirst ){
      for( i = 0; i < 256; i++ ){
	d = 0;
	if( i & 0x80 ) d |= 0x1;
	if( i & 0x40 ) d |= 0x2;
	if( i & 0x20 ) d |= 0x4;
	if( i & 0x10 ) d |= 0x8;
	if( i & 0x8 ) d |= 0x10;
	if( i & 0x4 ) d |= 0x20;
	if( i & 0x2 ) d |= 0x40;
	if( i & 0x1 ) d |= 0x80;
	table[i] = (unsigned char)d;
      }

      ptr = pmf->pm.ptr;
      for( size = PMFSize( pmf ); size > 0; size--, ptr++ )
	*ptr = table[*ptr];
    }
  }
}


static int
event_proc()
{
  XEvent   ev;
  PMFPixel pix;
  int      x, y;

  XNextEvent( disp, &ev );
  switch( ev.type ){
  case ButtonPress:
    if( ev.xbutton.button == 3 )
      return( 0 );
    x = ev.xbutton.x;
    y = ev.xbutton.y;
    pix = PMFGetPixel( pmf, x, y );
    switch( pmf->type ){
    case PMF_BITMAP:
      printf( "%d at (%d %d)\n", pix.bit, x, y );
      break;
    case PMF_GRAYMAP:
      printf( "%d at (%d %d)\n", pix.gray, x, y );
      break;
    case PMF_PIXELMAP:
      printf( "(%d %d %d) at (%d %d)\n", 
	      pix.pixel.r, pix.pixel.g, pix.pixel.b, x, y );
      break;
    }
    break;
  case Expose:
    XPutImage( disp, win, DefaultGC(disp, 0), ximage,
	       0, 0, 0, 0, width, height);
    break;
  }

  return( 1 );
}


void 
main( argc, argv )
     int   argc;
     char *argv[];
{
  extern XImage  *CreateStaticGrayImage(), *CreatePseudoColorImage(),
                 *CreateTrueColorImage();
  int             i;
  char            label[128];
  unsigned char  *im, *cim, r, g, b;


  get_args( argc, argv );

  disp = XOpenDisplay( dispstr );

  visual = DefaultVisual( disp, 0 );

  win = XCreateSimpleWindow( disp, RootWindow(disp, 0),
			     0, 0, width, height, 1, 0, 0 );

  sprintf( label, "%s(%dx%d,%d) XPMFVIEW",
	   (( NULL == filename )? "stdin" : filename), width, height, depth );
  XStoreName( disp, win, label );

  bitmap_fit( disp, pmf );

  switch( visual->class ){
  case StaticGray:
    if( ShowMessage )
      fprintf( stderr, "StaticGray visual class.\n" );
    ximage = CreateStaticGrayImage( disp, visual,
				    pmf->pm.ptr, width, height, depth );
    break;
  case PseudoColor:
    if( ShowMessage )
      fprintf( stderr, "PseudoColor visual class.\n" );
    ximage = CreatePseudoColorImage( disp, visual,
				     pmf->pm.ptr, width, height, depth );
    break;
  case TrueColor:
    if( ShowMessage )
      fprintf( stderr, "TrueColor visual class.\n" );
    ximage = CreateTrueColorImage( disp, visual,
				   pmf->pm.ptr, width, height, depth );
    break;
  }

  XSelectInput( disp, win, ButtonPressMask | ExposureMask );

  XMapWindow( disp, win );

  while( event_proc() );

  exit( 0 );
}
